/*
 * IEEE_754.java
 *
 * Creado el 29 de octubre de 2007, 18:20
 */


package computacion.practica2;

/**
 * El objetivo de esta prctica es ilustrar el manejo del estndar IEEE-754 de
 * representacin en binario (con doble precisin) de los nmeros en coma 
 * flotante. Puede encontrarse informacin detallada sobre este estndar en la
 * pgina web
 *   http://www.etsimo.uniovi.es/~antonio/uned/ieee754/
 * que adems incluye una "calculadora" para pasar de flotante a binario con 
 * la que puede verificarse la correccin de los mtodos de abajo.
 *
 * @author Vctor Jimnez
 *
 * MTODOS NUMRICOS
 * LICENCIATURA DE MATEMTICAS
 * UNIVERSIDAD DE MURCIA
 */
public class IEEE_754 {
    
    public static void main(String[] args) {        
        double flotante = -3567.82;
        
        System.out.println("");
        System.out.println("Traducimos " + flotante 
                + " a su expresin binaria:");
        System.out.println(flotanteAbinario(flotante));
        System.out.println("Traducimos ahora su expresin binaria a coma " 
                + "flotante y recibimos el nmero de partida:");
        System.out.println(binarioAflotante(flotanteAbinario(flotante)));     
        
        System.out.println("");
        System.out.println("Verificamos que el mtodo siguiente calcula"
                + " correctamente el primer nmero positivo,");
        System.out.println("que en JAVA se representa como Double.MIN_VALUE:");
        System.out.println("Double.MIN_VALUE = " + Double.MIN_VALUE);
        System.out.println("Nmero siguiente a 0: " + siguiente(0));
        
        System.out.println("");
        System.out.println("Verificamos que el mtodo anterior calcula"
                + " correctamente el mayor nmero positivo,");
        System.out.println("que en JAVA se representa como Double.MAX_VALUE:");
        System.out.println("Double.MAX_VALUE = " + Double.MAX_VALUE);
        System.out.println("Nmero anterior a " 
                + "Double.POSITIVE_INFINITY (+infinito): "
                + anterior(Double.POSITIVE_INFINITY));
                
        System.out.println("");
        System.out.println("Escribimos los 200 primeros nmeros positivos:");
        flotante = 0;        
        for (int i = 1; i <= 200; i++) {
            flotante = siguiente(flotante);
            System.out.println(flotante);
        }
        
        System.out.println("");
        System.out.println("Escribimos los 200 ltimos nmeros positivos:");
        flotante = Double.POSITIVE_INFINITY;        
        for (int i = 1; i <= 200; i++) {
            flotante = anterior(flotante);  
        }
        for (int i = 1; i <= 200; i++) {
            System.out.println(flotante);
            flotante = siguiente(flotante);            
        }        
    }
    
    /**
     * El mtodo flotanteAbinario traduce un nmero en coma flotante (es decir,
     * un double) a su representacin binaria de 64 bits de acuerdo con el 
     * estndar IEEE-754. Los tres bloques de dgitos (correspondientes a 
     * signo, mantisa y exponente) se expresan separados por barras verticales.
     */
    public static String flotanteAbinario(double flotante) {
        String binario;        
        int exp = 0;
        double mant = Math.abs(flotante);
        String signo;
        String exponente;
        String mantisa;
        int contador;
        boolean normalizado;
        
        if (flotante == 0) {
            return "0|00000000000|" +
                    "0000000000000000000000000000000000000000000000000000";
        } else if (flotante == Double.POSITIVE_INFINITY) {
            return "0|11111111111|" +
                    "0000000000000000000000000000000000000000000000000000";
        } else if (flotante == Double.NEGATIVE_INFINITY) {
            return "1|11111111111|" +
                    "0000000000000000000000000000000000000000000000000000";
        } else if (Double.isNaN(flotante)) {
            return "1|11111111111|" +
                    "1111111111111111111111111111111111111111111111111111";
        } else {        
            if (flotante > 0) {
                signo = "0";
            } else {
                signo = "1";
            }   
            
            if (mant >= 1) {
                while (mant >= 2) {
                    exp = exp + 1;
                    mant = mant / 2;
                }
            } else {
                while (mant < 1) {
                    exp = exp - 1;
                    mant = mant * 2.;
                }
            }
            
            if (exp >= -1022) {
                normalizado = true;
                mant = mant - 1;
                exp = exp + 1023;
            } else {
                normalizado = false;
                mant = mant * Math.pow(2, exp + 1022);
                exp = 0;
            }
            
            exponente = "";
            while (exp > 0) {
                exponente = (exp % 2) + exponente ;
                exp = exp / 2;
            }
            contador = exponente.length();
            for (int i = 1; i <= 11 - contador; i++) {
                exponente = "0" + exponente;
            }
                        
            mantisa = "";
            for (int i = 1; i <= 52; i++) {
                mant = mant * 2;
                mantisa = mantisa + ((int) mant);
                mant = mant - ((int) mant); 
            }            
           
            binario = signo + "|" + exponente + "|" + mantisa;                        
            return binario;
        }        
    
    } // fin del mtodo flotanteAbinario

    /**
     * El mtodo binarioAflotante traduce la representacin binaria de 
     * de un nmero conforme al estndar IEEE-754 a un nmero es coma flotante
     * (un double). Si la expresin introducida es incorrecta, es decir, 
     * no se adapta a la estructura "signo|exponente|mantisa" devuelve como
     * resultado Double.NaN (un no-nmero).
     */    
    public static double binarioAflotante (String binario) {
        String signo;
        String exponente;
        String mantisa;
        int exp;
        double flotante;
        int potencia = 1;
        
        if (binario.length() != 66) {
            System.out.println("La longitud de la cadena es incorrecta");
            return Double.NaN;
        }
        if ((binario.charAt(1) != '|') || (binario.charAt(13) != '|')) {
            System.out.println("No se han usado barras separadoras");
            return Double.NaN;
        }
        signo = binario.substring(0, 1);
        if ((signo.charAt(0) != '0') && (signo.charAt(0) != '1')) {
            System.out.println("Alguno de los caracteres es distinto de 0 y 1");
            return Double.NaN;
        }
        exponente = binario.substring(2, 13);
        for (int i = 0; i < 11; i++) {
            if ((exponente.charAt(i) != '0') && (exponente.charAt(i) != '1')) {
                System.out.println("Alguno de los caracteres es distinto de 0 y 1");
                return Double.NaN;
            }
        }        
        mantisa = binario.substring(14, 66);
        for (int i = 0; i < 52; i++) {
            if ((mantisa.charAt(i) != '0') && (mantisa.charAt(i) != '1')) {
                System.out.println("Alguno de los caracteres es distinto de 0 y 1");
                return Double.NaN;
            }
        }       
        if (exponente.equals("11111111111")){
            if (signo.equals("0") && mantisa.equals(
                    "0000000000000000000000000000000000000000000000000000")) {
                return Double.POSITIVE_INFINITY;
            } else if (signo.equals("1") && mantisa.equals(
                    "0000000000000000000000000000000000000000000000000000")) {
                return Double.NEGATIVE_INFINITY;
            } else {
                return Double.NaN;
            }
        }
 
        exp = 0;
        for (int i = 0; i <= 10; i++) {
            exp = exp + Integer.parseInt(exponente.substring(10 - i, 11 - i))
                    * potencia;
            potencia = potencia * 2;
        }
                
        if (exp == 0) {            
            flotante = 0;
            for (int i = 0; i <= 51; i++) {
                flotante = flotante + 
                        Integer.parseInt(mantisa.substring(i, i + 1))
                        * Math.pow(2, -1023 - i);
            }
        } else {
            flotante = Math.pow(2, exp - 1023);
            for (int i = 0; i <= 51; i++) {                
                flotante = flotante + 
                        Integer.parseInt(mantisa.substring(i, i + 1))
                        * Math.pow(2, exp - 1023 - i - 1);                
            }
        }
        
        if (signo.equals("1")) { 
            flotante = -flotante;
        }     
        
        return flotante;
    } // fin del mtodo binarioAflotante
    
    public static double siguiente(double flotante) {
        String binario = flotanteAbinario(flotante);
        String signo = binario.substring(0, 1);
        String exponente = binario.substring(2, 13);
        String mantisa = binario.substring(14, 66);
        int cont;
        
        if ((flotante == Double.POSITIVE_INFINITY) || Double.isNaN(flotante)) {
            return flotante;
        } else if (signo.equals("0") && mantisa.equals(
                "1111111111111111111111111111111111111111111111111111")) {
            cont = 10;
            while (exponente.substring(cont, cont + 1).equals("1")) {
                cont = cont - 1;
            }
            exponente = exponente.substring(0, cont) + "1";
            for (int i = cont + 1; i < 11; i++) {
                exponente = exponente + "0";
            }
            return binarioAflotante(signo + "|" + exponente + "|"
                    + "0000000000000000000000000000000000000000000000000000");
        } else if (signo.equals("0")) {
            cont = 51;
            while (mantisa.substring(cont, cont + 1).equals("1")) {
                cont = cont - 1;
            }
            mantisa = mantisa.substring(0, cont) + "1";
            for (int i = cont + 1; i < 52; i++) {
                mantisa = mantisa + "0";
            }
            return binarioAflotante(signo + "|" + exponente + "|" + mantisa);
        } else if (signo.equals("1") && mantisa.equals(
                "0000000000000000000000000000000000000000000000000000")) {
            cont = 10;
            while (exponente.substring(cont, cont + 1).equals("0")) {
                cont = cont - 1;
            }
            exponente = exponente.substring(0, cont) + "0";
            for (int i = cont + 1; i < 11; i++) {
                exponente = exponente + "1";
            }
            return binarioAflotante(signo + "|" + exponente + "|"
                    + "1111111111111111111111111111111111111111111111111111");
        } else {
            cont = 51;
            while (mantisa.substring(cont, cont + 1).equals("0")) {
                cont = cont - 1;
            }
            mantisa = mantisa.substring(0, cont) + "0";
            for (int i = cont + 1; i < 52; i++) {
                mantisa = mantisa + "1";
            }
            return binarioAflotante(signo + "|" + exponente + "|" + mantisa);
        }        
        
    } // fin del mtodo siguiente
    
    public static double anterior(double flotante) {
        return -siguiente(-flotante);        
    }    
}
