/*
 * Ejercicio9.java
 *
 * Creado el 6 de noviembre de 2007, 10:10
 */


package computacion.practica3;
import auxiliar.*;

/**
 * El motivo por el que la funcin f(x) = e^{-x} sen^2 x - x^3 tiene dos ceros 
 * es el siguiente. Claramente la funcin no tiene ceros negativos.
 * Por tanto basta probar que la funcin es positiva para un 
 * intervalo (0, a] y que es estrictamente decreciente a partir de a.  Es fcil
 * comprobar que a = 0.47 cumple estas condiciones porque:
 * -La funcin e^{-x} sen^2 x es mayor que (1 - x)(x - x^3/6)^2 en el intervalo
 *   (0,1] y esta segunda funcin es mayor que x^3 en (0,a];
 * -La funcin derivada f'(x) = e^{-x} (sen 2x - sen^2 x) - 3 x^2 es menor que
 *   e^{-x} - 3 x^2 y esta funcin es negativa a partir de a.
 *
 * Como es sabido, si para una raz c de una funcin f(x) se cumple que f'(c) es 
 * no nulo, entonces el mtodo de Newton para c es cuadrtico porque la funcin 
 *   g(x) = x - f(x)/f'(x) 
 * cumple que g'(c) = f(c)f''(c)/(f'(c))^2 = 0. Tras hacer operaciones, puede
 * comprobarse tambin que 
 *   g''(c) = f''(c)/f'(c). 
 * Si f'(c) = 0 entonces todava puede garantizarse la convergencia (lineal) del
 * mtodo. En efecto, supongamos para simplificar que c = 0. Entonces podemos 
 * escribir 
 *   f(x) = a x^n + o(x^n)
 * para un cierto entero positivo n > 1 (siendo n el orden de la primera 
 * derivada que no se anula), con lo que tambin
 *   f'(x) = a n x^{n-1} + o(x^{n-1}),
 *   f''(x) = a n (n-1) x^{n-2} + o(x^{n-2}).
 * De aqu resulta que 
 *   f(x)f''(x) = a^2 n (n-1) x^{2n-2} + o(x^{2n-2}),
 *   (f'(x))^2 =  a^2 n^2 x^{2n-2} + o(x^{2n-2}),
 * con lo que 
 *   g'(c) = (n - 1)/n < 1.
 * Por ejemplo, para la funcin f(x) = e^{-x} * sen^2 x - x^3 del ejercicio de 
 * abajo la primera derivada que no se anula en el punto c = 0 es la
 * segunda, con lo que habr convergencia lineal con ratio de convergencia 
 * |g'(c)| = 1/2,  como as atestiguarn los clculos numricos.
 *
 * El mtodo de Newton mltiple consiste en aplicar el mtodo de Newton 
 * convencional a la funcin F(x) = f(x)/f'(x), que en el caso de que c sea una 
 * raz mltiple de f, tiene a c como una raz simple.
 * En este caso, la funcin g(x) = x - f(x)/f'(x) que define habitualmente
 * el mtodo de Newton se convierte en
 *   G(x) = x - F(x)/F'(x).
 * Obsrvese que incluso en el caso de que c sea una raz simple f(c) el mtodo 
 * puede seguir aplicndose (seguimos teniendo G'(c) = 0) pero no mejora la 
 * convergencia cuadrtica del mtodo de Newton convencional, pues 
 * usando que F(x) = x - g(x) se tiene que
 *   F'(x) = 1 - g'(x),
 *   F''(x) = - g''(x), 
 * con lo que
 *   G''(c) = F''(c)/F'(c) = - g''(c)/(1 - g'(c)) = -f''(c)/f'(c) 
 * y por tanto
 *   |G''(c)| = |g''(c)|,
 * as que el ratio de convergencia (que es la mitad de ese valor)
 * es exactamente el mismo. Por ejemplo, en el ejercicio que nos ocupa se 
 * obtiene 
 *   0.5 * |f''(c2)| / |f'(c2)| = 3.573442833416733
 * como ratio de convergencia.
 *
 * @author Vctor Jimnez
 *
 * MTODOS NUMRICOS
 * LICENCIATURA DE MATEMTICAS
 * UNIVERSIDAD DE MURCIA
 */
public class Ejercicio9 {
    
    public static void main(String[] args) {
        double raiz1 = 0;
        double raiz2 = 0;
        double xini;
        double xfin;
        FuncionC2Ej9 f = new FuncionC2Ej9();
        
        try {
            System.out.println("");
            System.out.println("Clculo de las dos races con el mtodo de "
                    + "Newton:");
            System.out.println("Raiz c1 = 0:");
            raiz1 = MetodosIterativos.newton(f, -1, 1E-16, 100);
            System.out.println("Raiz positiva c2: ");
            raiz2 = MetodosIterativos.newton(f, 1, 1E-16, 100);
            System.out.println("");
            System.out.println("Clculo de las dos races con el mtodo de "
                    + "Newton mltiple:");  
            System.out.println("Raiz c1 = 0:");
            raiz1 = MetodosIterativos.newtonMultiple(f, -1, 1E-16, 100);
            System.out.println("Raiz positiva c2: ");
            raiz2 = MetodosIterativos.newtonMultiple(f, 1, 1E-16, 100);
        } catch(MetodosIterativos.ErrorMetodosIterativos err){
        }
        
        System.out.println("");
        System.out.println("En c1 = 0 el mtodo de Newton presenta " 
                + "convergencia lineal...");
        xini = -1;
        for (int i = 0; i <= 15; i++) {
            xfin = xini - f.eval(xini) / f.derivada(xini);            
            System.out.println("|x_" + (i + 1) + "| / "
                    + "|x_" + i + "| = "
                    + (Math.abs(xfin) / Math.abs(xini)) );
            xini = xfin;
        }
        System.out.println("...pero no cuadrtica:");
        xini = -1;
        for (int i = 0; i <= 15; i++) {
            xfin = xini - f.eval(xini) / f.derivada(xini);            
            System.out.println("|x_" + (i + 1) + "| / "
                    + "|x_" + i + "|^2 = "
                    + (Math.abs(xfin) / (Math.abs(xini) * Math.abs(xini))) );
            xini = xfin;
        }
        System.out.println("Sin embargo, el mtodo de Newton mltiple " 
                + "presenta convergencia cuadrtica en c1 = 0:");
        xini = -1;
        for (int i = 0; i <= 5; i++) {
            xfin = xini - (f.eval(xini) * f.derivada(xini)) /
                    (f.derivada(xini) * f.derivada(xini) -
                    f.eval(xini) * f.derivada2(xini) );
            System.out.println("|x_" + (i + 1) + "| / "
                    + "|x_" + i + "|^2 = "
                    + (Math.abs(xfin) / (Math.abs(xini) * Math.abs(xini))) );
            xini = xfin;
        }
        
        System.out.println("");
        System.out.println("En c2 el mtodo de Newton presenta " 
                + "convergencia cuadrtica...");
        xini = 1;
        for (int i = 0; i <= 6; i++) {
            xfin = xini - f.eval(xini) / f.derivada(xini);            
            System.out.println("|x_" + (i + 1) + " - c2| / "
                    + "|x_" + i + " - c2|^2 = "
                    + (Math.abs(xfin - raiz2) 
                    / (Math.abs(xini - raiz2) * Math.abs(xini - raiz2))) );
            xini = xfin;
        }
        System.out.println("...al igual que el mtodo de Newton mltiple:");
        xini = 1;
        for (int i = 0; i <= 7; i++) {
            xfin = xini - (f.eval(xini) * f.derivada(xini)) /
                    (f.derivada(xini) * f.derivada(xini) -
                    f.eval(xini) * f.derivada2(xini) );
            System.out.println("|x_" + (i + 1) + " - c2| / "
                    + "|x_" + i + " - c2|^2 = "
                    + (Math.abs(xfin - raiz2) 
                    / (Math.abs(xini - raiz2) * Math.abs(xini - raiz2))) );
            xini = xfin;
        }        
        System.out.println( "Ratio de convergencia cuadrtica = " 
                + (0.5 * Math.abs(f.derivada2(raiz2)) 
                / Math.abs(f.derivada(raiz2))) );
    }
    
}
